"""
CMasher Version
===============
Stores the different versions of the *CMasher* package.

"""


# %% VERSIONS
# Default/Latest/Current version
__version__ = "1.7.1"  # keep in sync with pyproject.toml



import numpy as np
from matplotlib.legend_handler import HandlerBase


# Define legend handler class for artists that use colormaps
class _HandlerColorPolyCollection(HandlerBase):
    # Override create_artists to create an AxesImage resembling a colormap
    def create_artists(
        self, legend, artist, xdescent, ydescent, width, height, fontsize, trans
    ):
        from matplotlib.image import AxesImage

        # Obtain the Axes object of this legend
        ax = legend.axes

        # Obtain the colormap of the artist
        cmap = artist.cmap

        # Create an AxesImage to contain the colormap with proper dimensions
        image = AxesImage(
            ax, cmap=cmap, transform=trans, extent=[xdescent, width, ydescent, height]
        )

        # Set the data of the image
        image.set_data(np.arange(cmap.N)[np.newaxis, ...])

        # Return the AxesImage object
        return [image]



"""
CMasher
=======
Scientific colormaps for making accessible, informative and *cmashing* plots.

"""


# %% IMPORTS AND DECLARATIONS
# CMasher imports
from . import app_usage, cm, utils
from .__version__ import __version__
from .cm import *
from .utils import *

# All declaration
__all__ = ["app_usage", "cm", "utils"]
__all__.extend(cm.__all__)
__all__.extend(utils.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



"""
CMasher Colormaps
=================
Holds all the different colormaps that are in *CMasher*.

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib.colors import ListedColormap as LC

# All declaration
__all__: list[str] = []


# %% GLOBALS
# Type aliases
CMAP_DCT = dict[str, LC]

# Initialize empty dict to hold colormaps in
cmap_d: CMAP_DCT = {}
cmap_cd: dict[str, CMAP_DCT] = {
    "sequential": {},
    "diverging": {},
    "cyclic": {},
    "qualitative": {},
    "misc": {},
}



# this file is generated via scripts/generate_known_cmap_types.py, do not edit manually

_CMASHER_BUILTIN_MAP_TYPES = {
    "amber": "sequential",
    "amethyst": "sequential",
    "apple": "sequential",
    "arctic": "sequential",
    "bubblegum": "sequential",
    "chroma": "sequential",
    "cmr.amber": "sequential",
    "cmr.amethyst": "sequential",
    "cmr.apple": "sequential",
    "cmr.arctic": "sequential",
    "cmr.bubblegum": "sequential",
    "cmr.chroma": "sequential",
    "cmr.copper": "cyclic",
    "cmr.cosmic": "sequential",
    "cmr.dusk": "sequential",
    "cmr.eclipse": "sequential",
    "cmr.ember": "sequential",
    "cmr.emerald": "sequential",
    "cmr.emergency": "cyclic",
    "cmr.fall": "sequential",
    "cmr.flamingo": "sequential",
    "cmr.freeze": "sequential",
    "cmr.fusion": "diverging",
    "cmr.gem": "sequential",
    "cmr.ghostlight": "sequential",
    "cmr.gothic": "sequential",
    "cmr.guppy": "diverging",
    "cmr.holly": "diverging",
    "cmr.horizon": "sequential",
    "cmr.iceburn": "diverging",
    "cmr.infinity": "cyclic",
    "cmr.jungle": "sequential",
    "cmr.lavender": "sequential",
    "cmr.lilac": "sequential",
    "cmr.neon": "sequential",
    "cmr.neutral": "sequential",
    "cmr.nuclear": "sequential",
    "cmr.ocean": "sequential",
    "cmr.pepper": "sequential",
    "cmr.pride": "diverging",
    "cmr.prinsenvlag": "diverging",
    "cmr.rainforest": "sequential",
    "cmr.redshift": "diverging",
    "cmr.sapphire": "sequential",
    "cmr.savanna": "sequential",
    "cmr.seasons": "cyclic",
    "cmr.seaweed": "diverging",
    "cmr.sepia": "sequential",
    "cmr.sunburst": "sequential",
    "cmr.swamp": "sequential",
    "cmr.torch": "sequential",
    "cmr.toxic": "sequential",
    "cmr.tree": "sequential",
    "cmr.tropical": "sequential",
    "cmr.viola": "diverging",
    "cmr.voltage": "sequential",
    "cmr.waterlily": "diverging",
    "cmr.watermelon": "diverging",
    "cmr.wildfire": "diverging",
    "copper": "cyclic",
    "cosmic": "sequential",
    "dusk": "sequential",
    "eclipse": "sequential",
    "ember": "sequential",
    "emerald": "sequential",
    "emergency": "cyclic",
    "fall": "sequential",
    "flamingo": "sequential",
    "freeze": "sequential",
    "fusion": "diverging",
    "gem": "sequential",
    "ghostlight": "sequential",
    "gothic": "sequential",
    "guppy": "diverging",
    "holly": "diverging",
    "horizon": "sequential",
    "iceburn": "diverging",
    "infinity": "cyclic",
    "jungle": "sequential",
    "lavender": "sequential",
    "lilac": "sequential",
    "neon": "sequential",
    "neutral": "sequential",
    "nuclear": "sequential",
    "ocean": "sequential",
    "pepper": "sequential",
    "pride": "diverging",
    "prinsenvlag": "diverging",
    "rainforest": "sequential",
    "redshift": "diverging",
    "sapphire": "sequential",
    "savanna": "sequential",
    "seasons": "cyclic",
    "seaweed": "diverging",
    "sepia": "sequential",
    "sunburst": "sequential",
    "swamp": "sequential",
    "torch": "sequential",
    "toxic": "sequential",
    "tree": "sequential",
    "tropical": "sequential",
    "viola": "diverging",
    "voltage": "sequential",
    "waterlily": "diverging",
    "watermelon": "diverging",
    "wildfire": "diverging",
}



# %% IMPORTS
# Built-in imports
import argparse
import os
import re
import sys
from importlib import import_module
from textwrap import dedent

# Package imports
import matplotlib as mpl
import numpy as np

import cmasher as cmr

# CMasher imports
from cmasher import __version__

# All declaration
__all__ = ["main"]


# %% GLOBALS
# Define set of packages with colormaps
cmap_pkgs = {"cmocean", "colorcet", "palettable"}

# Define main description of this package
main_desc = dedent(
    """
    CMasher: Scientific colormaps for making accessible, informative and
    'cmashing' plots"""
)[1:]
main_epilog = dedent(
    """
    This CLI-tool provides access to several of CMasher's utility functions.
    As several commands require a colormap object to work, all Python packages
    defined in the 'CMR_CMAP_PKGS' environment variable in addition to the
    following packages %s are attempted to be imported before any command is
    executed."""
)[1:] % (tuple(cmap_pkgs),)


def _get_main_desc(source):
    """
    Retrieves the main description of the provided object `source` and returns
    it.

    The main description is defined as the first paragraph of its docstring.

    Parameters
    ----------
    source : object
        The object whose main description must be retrieved.

    Returns
    -------
    main_desc : str or None
        The main description string of the provided `source` or *None* if
        `source` has not docstring.

    """
    # this function is copied from the e13tools package

    # Retrieve the docstring of provided source
    doc = source.__doc__

    # If doc is None, return None
    if doc is None:
        return None

    # Obtain the index of the last character of the first paragraph
    index = doc.find("\n\n")

    # If index is -1, there is only 1 paragraph
    if index == -1:
        index = len(doc)

    # Gather everything up to this index
    doc = doc[:index]

    # Replace all occurrences of 2 or more whitespace characters by a space
    doc = re.sub(r"\s{2,}", " ", doc)

    # Return doc
    return doc.strip()


# %% CLASS DEFINITIONS
# Define formatter that automatically extracts help strings of subcommands
class HelpFormatterWithSubCommands(argparse.ArgumentDefaultsHelpFormatter):
    # Override the add_argument function
    def add_argument(self, action):
        # Check if the help of this action is required
        if action.help is not argparse.SUPPRESS:
            # Check if this action is a subparser's action
            if isinstance(action, argparse._SubParsersAction):
                # If so, sort action.choices on name
                names = sorted(action.choices.keys())

                # Loop over all subcommands defined in the action
                for name in names:
                    # Obtain corresponding subparser
                    subparser = action.choices[name]

                    # Format the description of this subcommand and add it
                    self._add_item(
                        self.format_subcommands, [name, subparser.description]
                    )
            # Call super method in all other cases
            else:
                super().add_argument(action)

    # This function formats the description of a subcommand with given name
    def format_subcommands(self, name, description):
        # Determine the positions and widths of the help texts
        help_position = min(self._action_max_length + 2, self._max_help_position)
        help_width = max(self._width - help_position, 11)
        name_width = help_position - self._current_indent - 2

        # Transform name to the proper formatting
        name = "{0}{1: <{2}}{3}".format(
            " " * self._current_indent,
            name,
            name_width,
            "  " if (len(name) <= name_width) else "\n" + " " * help_position,
        )

        # Split the lines of the subcommand description
        desc_lines = self._split_lines(description, help_width)

        # Create list of all parts of the description of this subcommand
        parts = [name, desc_lines.pop(0), "\n"]

        # Loop over all remaining desc_lines
        for line in desc_lines:
            # Format and add to parts
            parts.append(f"{' '*help_position}{line}\n")

        # Convert to a single string and return
        return "".join(parts)


# %% COMMAND FUNCTION DEFINITIONS
# This function handles the 'bibtex' subcommand
def cli_bibtex():
    cmr.get_bibtex()


# This function handles the 'cmlist' subcommand
def cli_cmlist():
    # Obtain list without reversed versions
    cmaps = [
        cmap for cmap in cmr.get_cmap_list(ARGS.cmap_type) if not cmap.endswith("_r")
    ]

    # Create string with all cmaps
    cmaps_str = "\n".join(cmaps)

    # Print string
    print(cmaps_str)


# This function handles the 'cmtype' subcommand
def cli_cmap_type():
    # Import cmap packages
    import_cmap_pkgs()

    # Print cmap type
    print(cmr.get_cmap_type(get_cmap(ARGS.cmap)))


# This function handles the 'cmcolors' subcommand
def cli_cmap_colors():
    # Import cmap packages
    import_cmap_pkgs()

    # Obtain the colors
    colors = cmr.take_cmap_colors(
        get_cmap(ARGS.cmap),
        ARGS.ncolors,
        cmap_range=ARGS.cmap_range,
        return_fmt=ARGS.return_fmt,
    )

    # Print the colors line-by-line
    if ARGS.return_fmt in ("float", "norm"):
        np.savetxt(sys.stdout, colors, "%.8f")
    elif ARGS.return_fmt in ("int", "8bit"):
        np.savetxt(sys.stdout, colors, "%i")
    else:
        np.savetxt(sys.stdout, colors, "%s")


# This function handles the 'mkcmod' subcommand
def cli_mk_cmod():
    # Loop over all cmaps in ARGS.cmap
    for cmap in ARGS.cmap:
        # Create stand_alone module for this cmap
        cmap_path = cmr.create_cmap_mod(cmap, save_dir=ARGS.dir)

        # Print on commandline that module has been created
        print(f"Created standalone colormap module of {cmap} as {cmap_path}.")


# This function handles the 'cmview' subcommand
def cli_cmap_view():
    # Import cmap packages
    import_cmap_pkgs()

    # View cmap
    save = ARGS.cmap + ".png" if ARGS.save is True else ARGS.save
    cmr.view_cmap(
        get_cmap(ARGS.cmap), savefig=save, show_test=ARGS.test, show_grayscale=ARGS.gs
    )


# This function handles the 'app_usage tableau' subcommand
def cli_app_usage_tableau():
    # Create/Update Tableau properties file
    cmr.app_usage.update_tableau_pref_file(dirname=ARGS.dir)

    # Print on commandline that properties file was created/updated
    print(
        "Created/Updated Tableau preferences file in %r." % (os.path.abspath(ARGS.dir))
    )


# This function handles the 'lang_usage r' subcommand
def cli_lang_usage_r():
    # Print on commandline to look this up on the CMasher docs
    print(
        "Please see <https://cmasher.readthedocs.io/user/lang_usage/R.html> "
        "for instructions on how to access *CMasher* colormaps in *R*."
    )


# %% FUNCTION DEFINITIONS
# This function obtains the colormap that was requested
def get_cmap(cmap):
    # Try to obtain the colormap from MPL
    try:
        cmap = mpl.colormaps[cmap]

    # If this does not work, try to expand given cmap in setuptools-style
    except ValueError:
        # Check if cmap contains a colon
        if ":" in cmap:
            # Split cmap up into mod_name and obj_name
            mod_name, obj_name = cmap.split(":", 1)
            obj_path = obj_name.split(".")

            # Import the provided module as cmap
            cmap = import_module(mod_name)

            # Import the provided object from this module
            for obj in obj_path:
                cmap = getattr(cmap, obj)

    # If cmap is still a string, raise error
    if isinstance(cmap, str):
        # Print error and exit
        print(f"Requested 'CMAP' ({cmap!r}) cannot be found!")
        sys.exit()

    # Return cmap
    return cmap


# This function attempts to import a collection of packages with colormaps
def import_cmap_pkgs():
    # Obtain packages from CMR_CMAP_PKGS environment variable
    env_pkgs = os.environ.get("CMR_CMAP_PKGS", None)

    # Add env_pkgs to cmap_pkgs if it is not empty
    if env_pkgs is not None:
        # If Windows, split variable at semicolons
        if sys.platform.startswith("win"):
            env_pkgs = env_pkgs.split(";")
        # Else, if UNIX, split variable at colons
        elif sys.platform.startswith(("darwin", "linux")):
            env_pkgs = env_pkgs.split(":")
        # Else, ignore the variable
        else:
            env_pkgs = []

        # Add pkgs
        cmap_pkgs.update(env_pkgs)

    # Attempt to import each package
    for cmap_pkg in cmap_pkgs:
        try:
            import_module(cmap_pkg)
        except ImportError:
            pass


# This function adds the app_usage parser
def add_app_usage_parser(main_subparsers):
    # APP_USAGE COMMAND
    # Add app_usage parser
    parser = main_subparsers.add_parser(
        "app_usage",
        description=(
            "Access various function definitions useful for porting "
            "*CMasher* colormaps to other applications."
        ),
        formatter_class=HelpFormatterWithSubCommands,
        add_help=True,
    )

    # Set defaults for this parser
    parser.set_defaults(func="app_usage")

    # Create app_usage subparsers for app_usage parser
    subparsers = parser.add_subparsers(title="Applications", metavar="APP")

    # TABLEAU SUB-COMMAND
    # Add tableau subparser
    tableau_parser = subparsers.add_parser(
        "tableau",
        description=_get_main_desc(cmr.app_usage.update_tableau_pref_file),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Add 'dir' optional argument
    tableau_parser.add_argument(
        "-d",
        "--dir",
        help=(
            "Path to directory where the Tableau preferences file should be "
            "updated or created if it does not exist there."
        ),
        action="store",
        default=cmr.app_usage.update_tableau_pref_file.__defaults__[0],
        type=str,
    )

    # Set default for tableau parser
    tableau_parser.set_defaults(func=cli_app_usage_tableau)

    # Return parser
    return parser


# This function adds the lang_usage parser
def add_lang_usage_parser(main_subparsers):
    # LANG_USAGE COMMAND
    # Add lang_usage parser
    parser = main_subparsers.add_parser(
        "lang_usage",
        description=(
            "Access various function definitions useful for porting "
            "*CMasher* colormaps to other languages."
        ),
        formatter_class=HelpFormatterWithSubCommands,
        add_help=True,
    )

    # Set defaults for this parser
    parser.set_defaults(func="lang_usage")

    # Create lang_usage subparsers for lang_usage parser
    subparsers = parser.add_subparsers(title="Languages", metavar="LANG")

    # R SUB-COMMAND
    # Add tableau subparser
    r_parser = subparsers.add_parser(
        "r",
        description=(
            "Prints a URL to the online docs that describes how to "
            "access *CMasher* colormaps in *R*."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Set default for tableau parser
    r_parser.set_defaults(func=cli_lang_usage_r)

    # Return parser
    return parser


# %% MAIN FUNCTION
def main():
    """
    This is the main function of the CLI and is called whenever `cmr` is
    invoked from the command-line.

    """
    # Initialize argparser
    parser = argparse.ArgumentParser(
        "cmr",
        description=main_desc,
        epilog=main_epilog,
        formatter_class=HelpFormatterWithSubCommands,
        add_help=True,
        allow_abbrev=True,
    )

    # Add subparsers
    subparsers = parser.add_subparsers(title="commands", metavar="COMMAND")

    # OPTIONAL ARGUMENTS
    # Add 'version' argument
    parser.add_argument(
        "-v", "--version", action="version", version=f"CMasher v{__version__}"
    )

    # Create a cmap parser for several commands
    cmap_parent_parser = argparse.ArgumentParser(add_help=False)

    # Add 'cmap' argument
    cmap_parent_parser.add_argument(
        "cmap",
        help=(
            "Name of colormap to use as registered in *matplotlib* or the "
            "object path of a colormap (e.g., 'a.b:c.d' -> import a.b; "
            "cmap = a.b.c.d)."
        ),
        metavar="CMAP",
        action="store",
        type=str,
    )

    # BIBTEX COMMAND
    # Add bibtex subparser
    bibtex_parser = subparsers.add_parser(
        "bibtex",
        description=_get_main_desc(cmr.get_bibtex),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Set defaults for bibtex_parser
    bibtex_parser.set_defaults(func=cli_bibtex)

    # CMLIST COMMAND
    # Add cmlist subparser
    cmlist_parser = subparsers.add_parser(
        "cmlist",
        description=(
            "Prints a list of all colormaps available in *CMasher*. "
            "Reversed versions are excluded from this list."
        ),
        epilog=(
            "Note that *CMasher* colormaps registered in *MPL* have an "
            "added 'cmr.' prefix."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Add 'type' optional argument
    cmlist_parser.add_argument(
        "--type",
        help="Which colormap type to print.",
        action="store",
        default=cmr.get_cmap_list.__defaults__[0],
        choices=[
            "all",
            "s",
            "seq",
            "sequential",
            "d",
            "div",
            "diverging",
            "c",
            "cyc",
            "cyclic",
        ],
        type=str,
        dest="cmap_type",
    )

    # Set defaults for cmlist_parser
    cmlist_parser.set_defaults(func=cli_cmlist)

    # CMAP_TYPE COMMAND
    # Add cmap_type subparser
    cmap_type_parser = subparsers.add_parser(
        "cmtype",
        parents=[cmap_parent_parser],
        description=_get_main_desc(cmr.get_cmap_type),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Set defaults for cmap_type_parser
    cmap_type_parser.set_defaults(func=cli_cmap_type)

    # CMAP_COLORS COMMAND
    # Obtain the optional default arguments of take_cmap_colors
    defaults = cmr.take_cmap_colors.__kwdefaults__

    # Create a take_colors parser
    take_colors_parent_parser = argparse.ArgumentParser(add_help=False)

    # Add 'cmap_range' optional argument
    take_colors_parent_parser.add_argument(
        "--range",
        help=(
            "Normalized value range in the colormap from which colors should"
            " be taken."
        ),
        metavar=("LOWER", "UPPER"),
        action="store",
        nargs=2,
        default=defaults["cmap_range"],
        type=float,
        dest="cmap_range",
    )

    # Add 'fmt' optional argument
    take_colors_parent_parser.add_argument(
        "--fmt",
        help="Format to return colors in.",
        action="store",
        default=defaults["return_fmt"],
        choices=["float", "norm", "int", "8bit", "str", "hex"],
        type=str,
        dest="return_fmt",
    )

    # Add cmap_colors subparser
    cmap_colors_parser = subparsers.add_parser(
        "cmcolors",
        parents=[cmap_parent_parser, take_colors_parent_parser],
        description=_get_main_desc(cmr.take_cmap_colors),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Add 'N' argument
    cmap_colors_parser.add_argument(
        "ncolors",
        help="Number of colors to take.",
        metavar="N",
        action="store",
        type=int,
    )

    # Set defaults for cmap_colors_parser
    cmap_colors_parser.set_defaults(func=cli_cmap_colors)

    # RGB_TABLE COMMAND
    # Add rgb_table subparser
    rgb_table_parser = subparsers.add_parser(
        "rgbtable",
        parents=[cmap_parent_parser, take_colors_parent_parser],
        description="Retrieves the RGB values of the provided `cmap`.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Set defaults for rgb_table_parser
    rgb_table_parser.set_defaults(func=cli_cmap_colors, ncolors=None)

    # CMVIEW COMMAND
    # Add cmap_view subparser
    cmap_view_parser = subparsers.add_parser(
        "cmview",
        parents=[cmap_parent_parser],
        description=_get_main_desc(cmr.view_cmap),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Add 'save' optional argument
    cmap_view_parser.add_argument(
        "--save",
        help="Path where plot must be saved to.",
        metavar="PATH",
        action="store",
        nargs="?",
        default=None,
        const=True,
        type=str,
    )

    # Add 'test' optional argument
    cmap_view_parser.add_argument(
        "--test", help="Show a colormap test of `cmap` instead.", action="store_true"
    )

    # Add 'test' optional argument
    cmap_view_parser.add_argument(
        "--gs",
        "--grayscale",
        help="Show grayscale version of `cmap` as well.",
        action="store_true",
    )

    # Set defaults for cmap_view_parser
    cmap_view_parser.set_defaults(func=cli_cmap_view)

    # MK_CMOD COMMAND
    # Add mk_cmod subparser
    mk_cmod_parser = subparsers.add_parser(
        "mkcmod",
        description=_get_main_desc(cmr.create_cmap_mod),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    # Add 'cmap' argument
    mk_cmod_parser.add_argument(
        "cmap",
        help="Name of *CMasher* colormap(s) to create standalone module for.",
        metavar="CMAP",
        action="store",
        nargs="+",
        type=str,
    )

    # Add 'dir' optional argument
    mk_cmod_parser.add_argument(
        "-d",
        "--dir",
        help="Path to directory where the module must be saved.",
        action="store",
        default=cmr.create_cmap_mod.__kwdefaults__["save_dir"],
        type=str,
    )

    # Set defaults for mk_cmod_parser
    mk_cmod_parser.set_defaults(func=cli_mk_cmod)

    # APP_USAGE COMMAND
    # Add app_usage parser
    app_usage_parser = add_app_usage_parser(subparsers)

    # LANG_USAGE COMMAND
    # Add lang_usage parser
    lang_usage_parser = add_lang_usage_parser(subparsers)

    # Parse the arguments
    global ARGS
    ARGS = parser.parse_args()

    # If arguments is empty (no func was provided), show help
    if "func" not in ARGS:
        parser.print_help()
    # Else if 'func' is 'app_usage', show help of 'app_usage' parser
    elif ARGS.func == "app_usage":
        app_usage_parser.print_help()
    # Else if 'func' is 'lang_usage', show help of 'lang_usage' parser
    elif ARGS.func == "lang_usage":
        lang_usage_parser.print_help()
    # Else, call the corresponding function
    else:
        ARGS.func()



# %% IMPORTS
# Built-in imports
import os
from importlib.util import find_spec, module_from_spec, spec_from_file_location
from os import path

# Package imports
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.colors import ListedColormap as LC
from matplotlib.legend import Legend

# CMasher imports
import cmasher as cmr
from cmasher import cm as cmrcm
from cmasher.utils import (
    create_cmap_mod,
    create_cmap_overview,
    get_bibtex,
    get_cmap_list,
    get_sub_cmap,
    import_cmaps,
    set_cmap_legend_entry,
    take_cmap_colors,
    view_cmap,
)

HAS_VISCM = find_spec("viscm") is not None


def _MPL38_colormap_eq(cmap, other) -> bool:
    # equality check for two colormaps
    # this is adapted from Colormap.__eq__ (mpl 3.8)
    # in previous versions, colormaps compared as != unless they
    # had the exact same name, which is not what we care about here
    from matplotlib.colors import Colormap

    if not isinstance(other, Colormap) or cmap.colorbar_extend != other.colorbar_extend:
        return False
    # To compare lookup tables the Colormaps have to be initialized
    if not cmap._isinit:
        cmap._init()
    if not other._isinit:  # type: ignore [attr-defined]
        other._init()  # type: ignore [attr-defined]
    return np.array_equal(cmap._lut, other._lut)  # type: ignore [attr-defined]


# Save the path to this directory
dirpath = path.dirname(__file__)


# %% GLOBALS
# Obtain path to directory with colormaps
cmap_dir = path.abspath(path.join(dirpath, "../colormaps"))

# Obtain list of all colormaps defined in CMasher
# As all colormaps have their own directories, save them instead
_IGNORED = ["PROJECTS", "__pycache__"]
cm_names = [_ for _ in sorted(next(os.walk(cmap_dir))[1]) if _ not in _IGNORED]

# Obtain list of all colormaps registered in MPL
mpl_cmaps = plt.colormaps()


# %% PYTEST CLASSES AND FUNCTIONS
# Pytest class for create_cmap_mod
class Test_create_cmap_mod:
    # Test if a standalone module of rainforest can be created
    def test_standalone_rainforest(self):
        # Obtain the currently registered version of rainforest
        cmap_old = mpl.colormaps["cmr.rainforest"]

        # Create standalone module for rainforest
        cmap_path = create_cmap_mod("rainforest", _copy_name="rainforest_copy")

        # Try to import this module
        spec = spec_from_file_location("rainforest_copy", cmap_path)
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)

        # Check if the colormap in MPL has been updated
        cmap_new = mpl.colormaps["cmr.rainforest_copy"]

        # identity equality isn't achievable since mpl.colormaps.__getitem__
        # may return a copy
        assert cmap_new == mod.cmap

        if mpl.__version_info__ >= (3, 8):
            assert cmap_old == cmap_new
        else:
            assert _MPL38_colormap_eq(cmap_old, cmap_new)

        # Check if the values in both colormaps are the same
        assert np.allclose(cmap_old.colors, cmap_new.colors)

    # Test if a standalone module of infinity can be created
    def test_standalone_infinity(self):
        # Obtain the currently registered version of infinity
        cmap_old = mpl.colormaps["cmr.infinity"]

        # Create standalone module for infinity
        cmap_path = create_cmap_mod("infinity", _copy_name="inifinity_copy")

        # Try to import this module
        spec = spec_from_file_location("infinity_copy", cmap_path)
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)

        # Check if the colormap in MPL has been updated
        cmap_new = mpl.colormaps["cmr.infinity"]
        if mpl.__version_info__ >= (3, 8):
            assert cmap_old == cmap_new
        else:
            assert _MPL38_colormap_eq(cmap_old, cmap_new)

        assert cmap_old == cmap_new

        # Check if the values in both colormaps are the same
        assert np.allclose(cmap_old.colors, cmap_new.colors)

        # Check that the shifted version of infinity also exists
        assert "cmr.infinity_s" in plt.colormaps()

    # Test if providing an invalid colormap name fails
    def test_invalid_cmap(self):
        # Check if a ValueError is raised
        with pytest.raises(ValueError):
            create_cmap_mod("this is an incorrect colormap name")


# Pytest class for create_cmap_overview
class Test_create_cmap_overview:
    # Test if default arguments work
    def test_default(self):
        create_cmap_overview()

    # Test if providing a custom list of colormaps works
    def test_list_no_types(self):
        create_cmap_overview([cmrcm.rainforest], use_types=0, sort="lightness")
        create_cmap_overview(
            ["cmr.rainforest", "cmr.rainforest_r"], use_types=0, title=0
        )

    # Test if providing a custom list of colormaps works
    def test_list_cat(self):
        create_cmap_overview(["cmr.rainforest"])

    # Test if providing all MPL colormap objects works
    def test_mpl_cmaps_objs(self):
        cmaps = map(mpl.colormaps.__getitem__, mpl_cmaps)
        create_cmap_overview(cmaps, sort="perceptual")

    # Test if providing all MPL colormap names works
    def test_mpl_cmaps_names(self):
        create_cmap_overview(mpl_cmaps, sort="lightness")

    # Test if the lightness profiles can be plotted
    def test_lightness_profiles(self):
        create_cmap_overview(mpl_cmaps, plot_profile=True)

    # Test if dark mode can be enabled for the overview
    def test_dark_mode(self):
        create_cmap_overview([cmrcm.rainforest], dark_mode=True)

    # Test if the grayscale versions can be removed
    def test_no_grayscale(self):
        create_cmap_overview([cmrcm.rainforest], show_grayscale=False)

    # Test if the lightness/perceptual info can be shown
    def test_show_info(self):
        create_cmap_overview([cmrcm.rainforest], show_info=True)

    # Test if providing a custom dict of colormaps works
    def test_dict(self):
        create_cmap_overview({"test1": [cmrcm.rainforest], "test2": ["cmr.rainforest"]})

    # Test if the figure can be saved
    def test_savefig(self):
        create_cmap_overview(savefig="test.png")
        assert path.exists("./test.png")

    # test if providing an invalid sort value raises an error
    def test_invalid_sort_value(self):
        with pytest.raises(ValueError):
            create_cmap_overview(sort="test")


# Pytest for get_bibtex()-function
def test_get_bibtex():
    # Print the output of the get_bibtex() function
    get_bibtex()


# Pytest class for get_cmap_list()-function
class Test_get_cmap_list:
    # Test default arguments
    def test_default(self):
        assert get_cmap_list() == list(cmrcm.cmap_d)

    # Test obtaining all sequential colormaps
    def test_seq(self):
        assert get_cmap_list("seq") == list(cmrcm.cmap_cd["sequential"])

    # Test obtaining all diverging colormaps
    def test_div(self):
        assert get_cmap_list("div") == list(cmrcm.cmap_cd["diverging"])

    # Test obtaining all cyclic colormaps
    def test_cyc(self):
        assert get_cmap_list("cyc") == list(cmrcm.cmap_cd["cyclic"])


# Pytest class for get_sub_cmap()-function
class Test_get_sub_cmap:
    # Test if a copy can be made of the 'rainforest' colormap
    def test_rainforest_copy(self):
        assert np.allclose(
            get_sub_cmap("cmr.rainforest", 0, 1).colors, cmrcm.rainforest.colors
        )

    # Test if a sub can be made
    def test_rainforest_sub(self):
        sub_cmap = get_sub_cmap("cmr.rainforest", 0.2, 0.8)
        assert np.allclose(sub_cmap(0.0), cmrcm.rainforest(51))
        assert np.allclose(sub_cmap(1.0), cmrcm.rainforest(204))
        assert sub_cmap.N == 154

    # Test if a qualitative colormap can be made
    def test_lilac_qual(self):
        qual_cmap = get_sub_cmap("cmr.lilac", 0.2, 0.8, N=5)
        assert np.allclose(qual_cmap(0.0), cmrcm.lilac(51))
        assert np.allclose(qual_cmap(1.0), cmrcm.lilac(204))
        assert qual_cmap.N == 5

    # Test if providing an incorrect range raises an error
    def test_invalid_range(self):
        with pytest.raises(ValueError):
            get_sub_cmap("cmr.rainforest", -1, 1.5)


# Pytest class for import_cmaps()-function
class Test_import_cmaps:
    # Test if all colormaps in cmasher/colormaps are loaded into CMasher/MPL
    @pytest.mark.parametrize("cm_name", cm_names)
    def test_CMasher_cmaps(self, cm_name):
        # Check if provided cm_name is registered in CMasher and MPL
        for name in (cm_name, cm_name + "_r"):
            cmr_cmap = getattr(cmr, name)
            mpl_cmap = mpl.colormaps["cmr." + name]
            assert isinstance(cmr_cmap, LC)
            assert isinstance(mpl_cmap, LC)
            assert getattr(cmrcm, name) is cmr_cmap
            assert cmrcm.cmap_d.get(name) is cmr_cmap
            assert np.allclose(cmr_cmap.colors, mpl_cmap.colors)

    # Test if providing a cmap .txt-file works
    def test_cmap_file_txt(self):
        import_cmaps(
            path.join(dirpath, "../colormaps/cm_rainforest.txt"),
            _skip_registration=True,
        )

    # Test if providing a cmap .txt-file with 8-bit values works
    def test_cmap_file_8bit(self):
        import_cmaps(path.join(dirpath, "data/cm_8bit.txt"))

    # Test if providing a cmap .txt-file with HEX values works
    def test_cmap_file_HEX(self):
        import_cmaps(path.join(dirpath, "data/cm_hex.txt"))

    # Test if providing a cmap .jscm-file works
    @pytest.mark.skipif(not HAS_VISCM, reason="viscm is required")
    def test_cmap_file_jscm(self):
        cmap_path = path.join(dirpath, "data/cm_rainforest_jscm.jscm")
        import_cmaps(cmap_path)

    # Test if providing a cyclic colormap works
    def test_cyclic_cmap(self):
        name = "cyclic"
        import_cmaps(path.join(dirpath, f"data/cm_{name}.txt"))
        for cmap in [name, name + "_r", name + "_s", name + "_s_r"]:
            assert "cmr." + cmap in plt.colormaps()
            assert getattr(cmrcm, cmap, None) is not None
            assert cmap in cmrcm.cmap_d
            assert cmap in cmrcm.cmap_cd["cyclic"]

    # Test if providing a non-existing directory raises an error
    def test_non_existing_dir(self):
        with pytest.raises(OSError):
            import_cmaps("./test")

    # Test if providing an invalid cmap file raises an error
    def test_invalid_cmap_file(self):
        with pytest.raises(OSError):
            import_cmaps(path.join(dirpath, "data/test.txt"))

    # Test if providing an invalid cmap .npy-file raises an error
    def test_invalid_cmap_npy_file(self):
        with pytest.raises(ValueError):
            import_cmaps(path.join(dirpath, "data/cm_test2.npy"))

    # Test if providing a custom directory with invalid cmaps raises an error
    def test_invalid_cmap_dir(self):
        with pytest.raises(ValueError):
            import_cmaps(path.join(dirpath, "data"))


# Pytest class for set_cmap_legend_entry()-function
class Test_set_cmap_legend_entry:
    # Test if providing a PathCollection object works
    def test_path_collection(self):
        # Create a small scatter plot
        plot = plt.scatter([1, 2, 3], [2, 3, 4], c=[3, 4, 5], cmap=cmr.rainforest)

        # Add a cmap legend entry
        set_cmap_legend_entry(plot, "Test")

        # Check if the plot now has a special handler
        assert plot in Legend.get_default_handler_map()

        # Create a legend
        plt.legend()

        # Close the plot
        plt.close()

    # Test if providing a Line2D object does not work
    def test_line2d(self):
        # Create a small line plot
        plot = plt.plot([1, 2, 3], [2, 3, 4])

        # Attempt to add the cmap legend entry
        with pytest.raises(ValueError):
            set_cmap_legend_entry(plot, "Test")

        # Close the plot
        plt.close()


# Pytest class for take_cmap_colors()-function
class Test_take_cmap_colors:
    # Test if all colors can be taken from the 'rainforest' colormap
    def test_rainforest_all(self):
        assert np.allclose(
            take_cmap_colors("cmr.rainforest", None), cmr.rainforest.colors
        )

    # Test if five colors can be taken from the 'rainforest' colormap
    def test_rainforest_five(self):
        assert np.allclose(
            take_cmap_colors("cmr.rainforest", 5),
            [
                (0.0, 0.0, 0.0),
                (0.226123592, 0.124584033, 0.562997277),
                (0.0548210513, 0.515835251, 0.45667819),
                (0.709615979, 0.722863985, 0.0834727592),
                (1.0, 1.0, 1.0),
            ],
        )

    # Test if their 8-bit RGB values can be requested as well
    def test_rainforest_five_8bit(self):
        assert np.allclose(
            take_cmap_colors("cmr.rainforest", 5, return_fmt="int"),
            [(0, 0, 0), (58, 32, 144), (14, 132, 116), (181, 184, 21), (255, 255, 255)],
        )

    # Test if their HEX-code values can be requested as well
    def test_rainforest_five_hex(self):
        assert take_cmap_colors("cmr.rainforest", 5, return_fmt="hex") == [
            "#000000",
            "#3A2090",
            "#0E8474",
            "#B5B815",
            "#FFFFFF",
        ]

    # Test if only a subrange can be used for picking colors
    def test_rainforest_sub_five(self):
        assert take_cmap_colors(
            "cmr.rainforest", 5, cmap_range=(0.2, 0.8), return_fmt="hex"
        ) == ["#3E0374", "#10528A", "#0E8474", "#5CAD3C", "#D6BF4A"]

    # Test if providing an incorrect range raises an error
    def test_invalid_range(self):
        with pytest.raises(ValueError):
            take_cmap_colors("cmr.rainforest", 5, cmap_range=(-1, 1.5))


# Pytest class for view_cmap()-function
class Test_view_cmap:
    # Test if default arguments work
    def test_default(self):
        view_cmap("cmr.rainforest")

    # Test if the figure can be saved
    def test_savefig(self):
        view_cmap(
            "cmr.rainforest", show_test=True, show_grayscale=True, savefig="test.png"
        )
        assert path.exists("./test.png")
